# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
def test_slurm_time_to_seconds():
    from pybatch.tools import slurm_time_to_seconds as converter
    from pybatch import PybatchException

    assert converter(" ") == ""
    assert converter("10") == "600"
    assert converter("10:30") == "630"
    assert converter("100:30") == "6030"
    assert converter("2:10:5") == "7805"
    assert converter("2:10:05") == "7805"
    assert converter("2-2:10:30") == "180630"
    assert converter("2-2") == "180000"
    assert converter("2-2:10") == "180600"
    try:
        converter("2-0-4")
    except PybatchException:
        pass
    else:
        assert 0  # Exception expected
    try:
        converter("xvi")
    except PybatchException:
        pass
    else:
        assert 0  # Exception expected
    try:
        converter("1:2:3:4")
    except PybatchException:
        pass
    else:
        assert 0  # Exception expected



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
__all__ = [
    "GenericJob",
    "GenericProtocol",
    "LaunchParameters",
    "ConnectionParameters",
    "create_job",
    "PybatchException",
]
from .genericjob import GenericJob
from .generic_protocol import GenericProtocol
from .parameter import LaunchParameters, ConnectionParameters
from .job_factory import create_job

# WARNING `python_requires = >= 3.8`
from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "salome-pybatch"
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


class PybatchException(Exception):
    pass



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
from pathlib import Path
import os

from ... import GenericJob, GenericProtocol, LaunchParameters, PybatchException
from ...protocols.local import LocalProtocol
from ...tools import path_join, is_absolute, slurm_time_to_seconds, remote_mkdir


class Job(GenericJob):
    def __init__(self, param: LaunchParameters, protocol: GenericProtocol):
        self.job_params = param
        self.protocol: GenericProtocol
        if protocol is None:
            self.protocol = LocalProtocol()
        else:
            self.protocol = protocol
        self.jobid = ""
        self.remote_manager_path = path_join(
            param.work_directory, "pybatch_manager.py", is_posix=param.is_posix
        )

    def submit(self) -> None:
        try:
            logdir = path_join(
                self.job_params.work_directory,
                "logs",
                is_posix=self.job_params.is_posix,
            )
            remote_mkdir(self.protocol, logdir, self.job_params.python_exe)

            file_dir = Path(os.path.dirname(__file__))
            manager_script = file_dir / "pybatch_manager.py"
            input_files = self.job_params.input_files + [manager_script]
            self.protocol.upload(input_files, self.job_params.work_directory)
            command = [
                self.job_params.python_exe,
                self.remote_manager_path,
                "submit",
                self.job_params.work_directory,
            ]
            if self.job_params.wall_time:
                seconds = slurm_time_to_seconds(self.job_params.wall_time)
                command += ["--wall_time", seconds]
            if self.job_params.create_nodefile:
                if self.job_params.ntasks > 0:
                    command += ["--ntasks", str(self.job_params.ntasks)]
            if self.job_params.total_jobs > 1:
                command += ["--total_jobs", str(self.job_params.total_jobs)]
                if self.job_params.max_simul_jobs > 1:
                    command += [
                        "--max_simul_jobs",
                        str(self.job_params.max_simul_jobs),
                    ]
            command += self.job_params.command
            self.jobid = self.protocol.run(command).strip()
            int(self.jobid)  # check
        except Exception as e:
            message = "Failed to submit job."
            raise PybatchException(message) from e

    def wait(self) -> None:
        "Wait until the end of the job."
        if not self.jobid:
            return
        try:
            command = [
                self.job_params.python_exe,
                self.remote_manager_path,
                "wait",
                self.jobid,
            ]
            self.protocol.run(command)
        except Exception as e:
            message = "Failed to wait job."
            raise PybatchException(message) from e

    def state(self) -> str:
        if not self.jobid:
            return "CREATED"
        try:
            command = [
                self.job_params.python_exe,
                self.remote_manager_path,
                "state",
                self.jobid,
                self.job_params.work_directory,
            ]
            result: str = self.protocol.run(command).strip()

        except Exception as e:
            message = "Failed to wait job."
            raise PybatchException(message) from e
        return result

    def exit_code(self) -> int | None:
        exit_code_path = path_join(
            self.job_params.work_directory,
            "logs",
            "exit_code.log",
            is_posix=self.job_params.is_posix,
        )
        try:
            result = int(self.protocol.read(exit_code_path).strip())
        except Exception:
            result = None
        return result

    def cancel(self) -> None:
        if not self.jobid:
            return
        try:
            command = [
                self.job_params.python_exe,
                self.remote_manager_path,
                "cancel",
                self.jobid,
            ]
            self.protocol.run(command)
        except Exception as e:
            message = "Failed to cancel job."
            raise PybatchException(message) from e

    def get(self, remote_paths: list[str], local_path: str | Path) -> None:
        """Copy a file or directory from the remote work directory.

        :param remote_path: path relative to work directory on the remote host.
        :param local_path: destination of the copy on local file system.
        """
        checked_paths = []
        for path in remote_paths:
            if is_absolute(path, self.job_params.is_posix):
                checked_paths.append(path)
            else:
                p = path_join(
                    self.job_params.work_directory,
                    path,
                    is_posix=self.job_params.is_posix,
                )
                checked_paths.append(p)
        self.protocol.download(checked_paths, local_path)

    def stdout(self) -> str:
        output_file = path_join(
            self.job_params.work_directory,
            "logs",
            "output.log",
            is_posix=self.job_params.is_posix,
        )
        return self.protocol.read(str(output_file))

    def stderr(self) -> str:
        output_file = path_join(
            self.job_params.work_directory,
            "logs",
            "error.log",
            is_posix=self.job_params.is_posix,
        )
        return self.protocol.read(str(output_file))

    def batch_file(self) -> str:
        return ""



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
import tempfile
import os
import shutil
import sys
from pathlib import Path
import pybatch
import pybatch.protocols.local


def test_protocol_local() -> None:
    py_exe = sys.executable
    # Test configuration
    test_dir = tempfile.mkdtemp(suffix="_pybatchtest")
    work_dir = os.path.join(test_dir, "remote_dir")
    local_work_dir = os.path.join(test_dir, "local_dir")
    os.mkdir(work_dir)
    os.mkdir(local_work_dir)
    p = pybatch.protocols.local.LocalProtocol()
    test_file_name = "local_test.txt"
    remote_test_file = os.path.join(work_dir, test_file_name)
    local_test_file = os.path.join(local_work_dir, test_file_name)
    # Test download a remote path that does not exist.
    try:
        p.download([remote_test_file], local_test_file)
    except pybatch.PybatchException as e:
        assert remote_test_file in str(e)
    else:
        assert 0

    ## Test create a file in an inaccessible place
    try:
        p.create("/no/directory/", "file content")
    except FileNotFoundError as e:
        assert "directory" in str(e)
    else:
        assert 0

    # Test create + download
    file_content = "Servus!"
    p.create(remote_test_file, file_content)
    remote_content = p.read(remote_test_file)
    assert remote_content == file_content
    p.download([remote_test_file], local_test_file)
    assert Path(local_test_file).read_text() == file_content

    # download in an inaccessible place
    local_wrong_path = os.path.join(local_work_dir, "nodir", "nofile")
    try:
        p.download([remote_test_file], local_wrong_path)
    except FileNotFoundError as e:
        assert "nodir" in str(e)
    else:
        assert 0

    # upload nonexistent file
    try:
        p.upload([local_wrong_path], work_dir)
    except pybatch.PybatchException as e:
        assert "nodir" in str(e)
    else:
        assert 0

    # upload file to an inaccessible place
    try:
        p.upload([local_test_file], "/no/directory/")
    except FileNotFoundError as e:
        assert "directory" in str(e)
    else:
        assert 0

    # upload + download + check
    name_bis = "local_test_bis.txt"
    remote_test_file_bis = os.path.join(work_dir, name_bis)
    local_test_file_bis = os.path.join(local_work_dir, name_bis)
    p.upload([local_test_file], remote_test_file_bis)
    assert Path(remote_test_file_bis).read_text() == file_content
    p.download([remote_test_file_bis], local_test_file_bis)
    assert Path(local_test_file_bis).read_text() == file_content

    # run
    command = [py_exe, "-c", 'print("Cool!")']
    res = p.run(command)
    assert res.strip() == "Cool!"

    ## run error
    command = [py_exe, "-c", "exit(1)"]
    try:
        res = p.run(command)
    except pybatch.PybatchException as e:
        assert "Error 1" in str(e)
    else:
        assert 0

    shutil.rmtree(test_dir)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
"""Run a command with a supervisor."""

from __future__ import annotations
from typing import Optional
from types import FrameType
from pathlib import Path

import sys
import subprocess
import signal
import functools
import traceback


def handler(
    proc: subprocess.Popen[bytes], signum: int, frame: Optional[FrameType]
) -> None:
    proc.terminate()


def run() -> None:
    wall_time = int(sys.argv[1])
    if wall_time < 1:
        wall_time = None  # type: ignore
    command = sys.argv[2:]
    # daemonize process

    log_path = Path("logs")
    log_path.mkdir(parents=True, exist_ok=True)
    stdout_log = log_path / "output.log"
    stderr_log = log_path / "error.log"
    # file descriptors are automaticaly closed by default
    # (see close_fds argument of Popen).
    stdout_file = open(stdout_log, "w")
    stderr_file = open(stderr_log, "w")
    proc = subprocess.Popen(
        command,
        stdout=stdout_file,
        stderr=stderr_file,
    )
    signal.signal(signal.SIGTERM, functools.partial(handler, proc))
    try:
        exit_code = proc.wait(wall_time)
    except subprocess.TimeoutExpired:
        proc.terminate()
        exit_code = proc.wait()
    exit_log = log_path / "exit_code.log"
    with open(exit_log, "w") as exit_file:
        exit_file.write(str(exit_code))


if __name__ == "__main__":
    try:
        run()
    except Exception:
        error_log = Path("logs") / "error.log"
        error_message = traceback.format_exc()
        with open(error_log, "a") as logfile:
            logfile.write(error_message)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
from collections.abc import Iterable
from pathlib import Path
from ..parameter import ConnectionParameters
from ..tools import run_check, escape_str
from .. import PybatchException


class SshProtocol:
    """Communication protocol using the external commands ssh and scp."""

    def __init__(self, params: ConnectionParameters):
        self._host = params.host
        self._user = params.user
        self._password = params.password  # TODO not supported yet
        self._gss_auth = params.gss_auth

    def upload(
        self, local_entries: Iterable[str | Path], remote_path: str
    ) -> None:
        # conversion Path to str for mypy
        full_command = ["scp", "-r"] + list(str(x) for x in local_entries)
        destination = ""
        if self._user:
            destination += self._user + "@"
        destination += self._host + ":" + remote_path
        full_command.append(escape_str(destination))
        run_check(full_command)

    def download(
        self, remote_entries: Iterable[str], local_path: str | Path
    ) -> None:
        command = ["scp", "-r"]
        remote_id = ""
        if self._user:
            remote_id += self._user + "@"
        remote_id += self._host + ":"
        for entry in remote_entries:
            full_command = command + [
                escape_str(remote_id + entry),
                str(local_path),
            ]
            run_check(full_command)

    def create(self, remote_path: str, content: str) -> None:
        full_command = ["ssh", "-T", self._host]
        if self._user:
            full_command += ["-l", self._user]
        if self._gss_auth:
            full_command.append("-K")
        full_command.append(f"cat > '{remote_path}'")
        run_check(full_command, input=content)

    def read(self, remote_path: str) -> str:
        full_command = ["ssh", self._host]
        if self._user:
            full_command += ["-l", self._user]
        if self._gss_auth:
            full_command.append("-K")
        full_command.append(f"cat '{remote_path}'")
        proc = run_check(full_command)
        return proc.stdout

    def run(self, command: list[str]) -> str:
        if len(command) == 0:
            raise PybatchException("Empty command.")
        full_command = ["ssh", self._host]
        if self._user:
            full_command += ["-l", self._user]
        if self._gss_auth:
            full_command.append("-K")
        full_command.append(command[0])
        for arg in command[1:]:
            full_command.append(escape_str(arg))
        proc = run_check(full_command)
        return proc.stdout


def open(params: ConnectionParameters) -> SshProtocol:
    return SshProtocol(params)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
import pathlib
import subprocess
import typing
from . import PybatchException
from .generic_protocol import GenericProtocol


def path_join(base: str, *paths: str, is_posix: bool) -> str:
    result: pathlib.PurePath
    if is_posix:
        result = pathlib.PurePosixPath(base)
    else:
        result = pathlib.PureWindowsPath(base)
    for path in paths:
        result = result / path
    return str(result)


def is_absolute(path: str, is_posix: bool) -> bool:
    if is_posix:
        return pathlib.PurePosixPath(path).is_absolute()
    else:
        return pathlib.PureWindowsPath(path).is_absolute()


def slurm_time_to_seconds(val: str) -> str:
    """Convert a slurm time format string to seconds.

    See https://slurm.schedmd.com/sbatch.html#OPT_time
    Acceptable time formats:
       "minutes", "minutes:seconds", "hours:minutes:seconds",
       "days-hours", "days-hours:minutes" and "days-hours:minutes:seconds"
    """
    val = val.strip()
    if not val:
        return val
    try:
        day_split = val.split("-")
        if len(day_split) == 2:
            days = int(day_split[0])
            rem = day_split[1]
        elif len(day_split) == 1:
            days = 0
            rem = day_split[0]
        else:
            raise PybatchException(f"Invalid time format: {val}.")
        hour_split = rem.split(":")
        if len(hour_split) == 3:
            hours = int(hour_split[0])
            minutes = int(hour_split[1])
            seconds = int(hour_split[2])
        elif len(hour_split) == 2:
            if days > 0:  # days-hours:minutes
                hours = int(hour_split[0])
                minutes = int(hour_split[1])
                seconds = 0
            else:  # minutes:seconds
                hours = 0
                minutes = int(hour_split[0])
                seconds = int(hour_split[1])
        elif len(hour_split) == 1:
            if days > 0:  # days-hours
                hours = int(hour_split[0])
                minutes = 0
                seconds = 0
            else:  # minutes
                hours = 0
                minutes = int(hour_split[0])
                seconds = 0
        else:
            raise PybatchException(f"Invalid time format: {val}.")
    except Exception as e:
        raise PybatchException(f"Invalid time format: {val}.") from e
    result = seconds + 60 * minutes + 3600 * hours + 24 * 3600 * days
    return str(result)


def run_check(
    command: list[str], **extra: typing.Any
) -> subprocess.CompletedProcess[str]:
    proc = subprocess.run(command, capture_output=True, text=True, **extra)
    ret_code = proc.returncode
    if ret_code != 0:
        message = f"""Error {ret_code}.
  command: {command}.
  stderr: {proc.stderr}
"""
        raise PybatchException(message)
    return proc


def escape_str(val: str) -> str:
    """Escape characters with special meaning in bash.
    a'b -> 'a'\''b'
    a b -> 'a b'
    """
    special_chars = " ()[]{}*?$#'\\"
    special_found = False
    for c in special_chars:
        if c in val:
            special_found = True
            break
    if special_found:
        result = "'" + val.replace("'", "'\\''") + "'"
    else:
        result = val
    return result


def remote_mkdir(protocol: GenericProtocol, dir: str, python_exe: str) -> None:
    """Create a directory on a remote server.

    The directory is created by running a python command.
    :param protocol: Connection protocol to the remote server.
    :param dir: Path of the remote directory to be created.
    :param python_exe: Path to the python executable on the remote server.
    """
    py_script = f"from pathlib import Path; Path({repr(dir)}).mkdir(parents=True, exist_ok=True)"
    command = [python_exe, "-c", py_script]
    protocol.run(command)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
"""Local tests for any plugin."""

import tempfile
from pathlib import Path
import os
import shutil
import time
import sys


def test_python_script(job_plugin):
    """Launch a python script which ends without errors.

    Check files from 'logs' folder.
    """
    import pybatch

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "hello.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "hello.py", "world"],
        workdir,
        input_files=[script],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    job.submit()
    job.wait()

    assert "Hello world !" in job.stdout()
    assert not job.stderr()
    assert job.exit_code() == 0
    shutil.rmtree(workdir)


def test_finish_without_wait(job_plugin):
    "Test that a job can finish without using job.wait()."
    import pybatch

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "hello.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "hello.py", "world"],
        workdir,
        input_files=[script],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    job.submit()
    time.sleep(2)  # instead of job.wait()
    assert job.state() == "FINISHED"
    shutil.rmtree(workdir)


def test_error_script(job_plugin):
    """Launch a python script which ends in error.

    Check files from 'logs' folder.
    """
    import pybatch

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "error.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "error.py"],
        workdir,
        input_files=[script],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    job.submit()
    job.wait()

    assert "Problems comming..." in job.stdout()
    assert "Oups!" in job.stderr()
    assert job.exit_code() == 1
    shutil.rmtree(workdir)


def test_state(job_plugin):
    """Test the state of a job."""
    import pybatch

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "sleep.py", "1"],
        workdir,
        input_files=[script],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    assert job.state() == "CREATED"
    job.submit()
    assert job.state() == "RUNNING"
    job.wait()
    assert job.state() == "FINISHED"
    assert job.exit_code() == 0
    assert (Path(workdir) / "wakeup.txt").exists()
    shutil.rmtree(workdir)


def test_cancel(job_plugin):
    """Cancel a running job."""
    import pybatch
    import time

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "sleep.py", "2"],
        workdir,
        input_files=[script],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    assert job.state() == "CREATED"
    job.submit()
    time.sleep(1)
    job.cancel()
    job.wait()
    time.sleep(2)  # sleep.py would have had the time to finish if not canceled.
    assert not (Path(workdir) / "wakeup.txt").exists()
    assert job.exit_code() != 0
    shutil.rmtree(workdir)


def test_serialization(job_plugin):
    """Serialization / deserialization of a submited job in the same script."""
    import pybatch
    import pickle

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "sleep.py", "1"],
        workdir,
        input_files=[script],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    job.submit()
    pick_job = pickle.dumps(job)
    new_job = pickle.loads(pick_job)
    assert new_job.state() == "RUNNING"
    new_job.wait()
    assert new_job.state() == "FINISHED"
    shutil.rmtree(workdir)


def test_wall_time(job_plugin):
    """Job with wall time."""
    import pybatch

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "sleep.py", "3"],
        workdir,
        input_files=[script],
        wall_time="0:1",  # 1s
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    job.submit()
    job.wait()
    assert not (Path(workdir) / "wakeup.txt").exists()
    assert job.exit_code() != 0  # return value is OS specific
    shutil.rmtree(workdir)


def test_files_and_directories(job_plugin):
    """Job that uses and produces files and directories."""
    import pybatch

    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    resultdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "code.py"
    file_input = Path(current_file_dir) / "data" / "input.txt"
    dir_input = Path(current_file_dir) / "data" / "data"
    is_posix = os.name == "posix"
    params = pybatch.LaunchParameters(
        [sys.executable, "code.py", "1"],
        workdir,
        input_files=[script, file_input, dir_input],
        python_exe=sys.executable,
        is_posix=is_posix,
    )
    job = pybatch.create_job(job_plugin, params)
    job.submit()
    job.wait()
    assert not job.stderr()
    assert job.exit_code() == 0
    job.get(["output.txt"], resultdir)
    assert (Path(resultdir) / "output.txt").read_text() == "51"
    job.get(["data"], resultdir)
    assert (Path(resultdir) / "data" / "output.txt").read_text() == "69"
    shutil.rmtree(workdir)
    shutil.rmtree(resultdir)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
import typing


class GenericJob(typing.Protocol):
    "Job protocol to be implemented."

    def submit(self) -> None:
        """Submit the job to the batch manager and return.

        If the submission fails, raise an exception.
        """
        ...

    def wait(self) -> None:
        "Wait until the end of the job."
        ...

    def state(self) -> str:
        """Possible states : 'CREATED', 'IN_PROCESS', 'QUEUED', 'RUNNING',
        'PAUSED', 'FINISHED', 'FAILED'
        """
        ...

    def exit_code(self) -> int | None:
        """Get the exit code of the command if any.
        If the code is not found, for instance when the job is neither FINISHED
        nor FAILED, return None.
        """
        ...

    def cancel(self) -> None:
        "Stop the job."
        ...

    def get(self, remote_path: list[str], local_path: str) -> None:
        """Copy files from the remote work directory.

        :param remote_paths: paths relative to work directory on remote host.
        :param local_path: destination of the copy on local file system.
        """
        ...

    def batch_file(self) -> str:
        """Get the content of the batch file submitted to the batch manager.

        This function is not relevant if the job is not submitted to a batch
        manager that uses a submission file.
        """
        ...

    def stdout(self) -> str:
        "Standard output of the job."
        ...

    def stderr(self) -> str:
        "Standard error of the job."
        ...

    def dump(self) -> str:
        """Serialization of the job in a humanly readable format.

        This feature is not implemented yet because it is possible to use pickle
        for automatic serialization.
        """
        ...



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
import typing
from collections.abc import Iterable
from pathlib import Path
import shutil
import os

from .. import PybatchException
from ..tools import run_check


def copy(src: str | Path, dest: str | Path) -> None:
    """Recursively copy files and directories."""
    if os.path.isfile(src):
        shutil.copy(src, dest)
    elif os.path.isdir(src):
        src_basename = os.path.basename(src)
        dest_dir = Path(dest) / src_basename
        shutil.copytree(src, dest_dir, dirs_exist_ok=True)
    else:
        raise PybatchException(
            f"Copy error. Path {src} is neither a file, nor a directory."
        )


class LocalProtocol:
    "Protocol for localhost."

    def __init__(self, params: typing.Any = None):
        pass

    def upload(
        self, local_entries: Iterable[str | Path], remote_path: str
    ) -> None:
        for entry in local_entries:
            copy(entry, remote_path)

    def download(
        self, remote_entries: Iterable[str], local_path: str | Path
    ) -> None:
        for entry in remote_entries:
            copy(entry, local_path)

    def create(self, remote_path: str, content: str) -> None:
        Path(remote_path).write_text(content)

    def read(self, remote_path: str) -> str:
        return Path(remote_path).read_text()

    def run(self, command: list[str]) -> str:
        proc = run_check(command)
        return proc.stdout


def open() -> LocalProtocol:
    return LocalProtocol()



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#

# type: ignore
from __future__ import annotations
from pybatch import GenericJob, LaunchParameters, GenericProtocol
from .job import Job


class Plugin:
    @property
    def description(self) -> str:
        return """Job submission using slurm batch manager."""

    def create_job(
        self, param: LaunchParameters, connection_protocol: GenericProtocol
    ) -> GenericJob:
        return Job(param, connection_protocol)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path


@dataclass
class LaunchParameters:
    """Parameters of a job to be launched.

    List of parameters :

      * command - full command to run with arguments as a list.
      * work_directory - remote work directory, with path separator at the end.
      * host - remote host where the job will be launched.
      * user - user name if needed.
      * name - name of the job.
      * total_jobs - total number of jobs using a job array. The job index is
        passed as the last argument to the command.
      * max_simul_jobs - maximum number of simultaneous jobs using a job array.
      * nodes - number of required nodes, 0 for undefined.
      * ntasks - number of required tasks, 0 for undefined.
      * exclusive - activate exclusive mode.
      * wall_time - maximum time of the job.
        Acceptable time formats include "minutes", "minutes:seconds",
        "hours:minutes:seconds", "days-hours", "days-hours:minutes" and
        "days-hours:minutes:seconds".
      * mem_per_node - memory required per node (ex. "32G").
      * mem_per_cpu - minimum memory required per usable allocated CPU.
      * queue - required queue.
      * partition - required partition.
      * wckey
      * extra_as_string - extra parameters as a string
        (ex. "#SBATCH --cpus-per-task=4").
      * extra_as_list - extra parameters as a list (ex. ["--cpus-per-task=4"]).
        job.
      * input_files - list of local files to be copied to remote work_directory.
      * is_posix - Unix like server (True) or Windows server (False).
      * python_exe - path to the python executable. Default to "python3".
      * create_nodefile - create LIBBATCH_NODEFILE which contains the list of
        allocated nodes.
    """

    command: list[str]
    work_directory: str
    name: str = ""
    total_jobs: int = 1
    max_simul_jobs: int = 1
    nodes: int = 0
    ntasks: int = 0
    exclusive: bool = False
    wall_time: str = ""
    mem_per_node: str = ""
    mem_per_cpu: str = ""
    queue: str = ""
    partition: str = ""
    wckey: str = ""
    extra_as_string: str = ""
    extra_as_list: list[str] = field(default_factory=list)
    input_files: list[str | Path] = field(default_factory=list)
    is_posix: bool = True
    python_exe: str = "python3"
    create_nodefile: bool = False


@dataclass
class ConnectionParameters:
    """Parameters needed to connect to a remote server.

    List of parameters :

    * host
    * user
    * password
    * gss_auth - use the gss api for authentication. It has to be True when
      using Kerberos protocole.
    """

    host: str = ""
    user: str | None = None
    password: str | None = None
    gss_auth: bool = False



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
import typing
from typing import Optional
from types import FrameType

from ... import GenericJob, LaunchParameters, PybatchException
from ...tools import slurm_time_to_seconds
from pathlib import Path
import shutil
import subprocess
import psutil
import json
import os
import signal
import socket
import functools
import sys


def handler(
    proc: subprocess.Popen[bytes], signum: int, frame: Optional[FrameType]
) -> None:
    proc.terminate()


def copy(src: str | Path, dest: str | Path) -> None:
    """Recursively copy files and directories."""
    if os.path.isfile(src):
        shutil.copy(src, dest)
    elif os.path.isdir(src):
        src_basename = os.path.basename(src)
        dest_dir = Path(dest) / src_basename
        shutil.copytree(src, dest_dir, dirs_exist_ok=True)
    else:
        raise PybatchException(
            f"Path {src} is neither a file, nor a directory."
        )


class Job(GenericJob):
    def __init__(self, param: LaunchParameters):
        self.command = param.command
        self.work_directory = param.work_directory
        self.input_files = param.input_files
        seconds_str = slurm_time_to_seconds(param.wall_time)
        if not seconds_str:
            self.wall_time = None
        else:
            self.wall_time = int(seconds_str)
        self.pid = -1  # job not launched
        self.ntasks = 0
        if param.create_nodefile:
            self.ntasks = param.ntasks
        # TODO We could also use ntasks and mem_per_node to set
        # limits using resource module.

    def submit(self) -> None:
        """Submit the job to the batch manager and return.

        The command is executed in a process detached from the current one, in
        a daemon mode.
        """
        self._prepare_run()
        if hasattr(os, "fork"):
            self._posix_submit()
        else:
            self._windows_submit()

    def wait(self) -> None:
        "Wait until the end of the job."
        if self.pid > 0:
            try:
                pu = psutil.Process(self.pid)
                pu.wait()
            except psutil.NoSuchProcess:
                pass

    def state(self) -> str:
        """Possible states : 'CREATED', 'IN_PROCESS', 'QUEUED', 'RUNNING',
        'PAUSED', 'FINISHED', 'FAILED'
        """
        if self.pid == -1:
            return "CREATED"
        if psutil.pid_exists(self.pid):
            return "RUNNING"
        else:
            exit_log = Path(self.work_directory) / "logs" / "exit_code.log"
            if exit_log.is_file():
                exit_value = exit_log.read_text()
                if exit_value == "0":
                    return "FINISHED"
                else:
                    return "FAILED"
            else:
                return "FAILED"

    def exit_code(self) -> int | None:
        exit_log = Path(self.work_directory) / "logs" / "exit_code.log"
        result = None
        try:
            if exit_log.is_file():
                exit_value = exit_log.read_text().strip()
                result = int(exit_value)
        except Exception:
            result = None
        return result

    def cancel(self) -> None:
        "Stop the job."
        pu = psutil.Process(self.pid)
        children = pu.children(recursive=True)
        for p in children:
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                pass
        pu.terminate()

    def get(self, remote_paths: list[str], local_path: str | Path) -> None:
        """Copy a file or directory from the remote work directory.

        :param remote_path: path relative to work directory on the remote host.
        :param local_path: destination of the copy on local file system.
        """
        for path in remote_paths:
            if os.path.isabs(path):
                abs_remote_path = path
            else:
                tmp_path = Path(self.work_directory) / path
                abs_remote_path = os.path.realpath(tmp_path)
            copy(abs_remote_path, local_path)

    def stdout(self) -> str:
        output_file = Path(self.work_directory, "logs", "output.log")
        return output_file.read_text()

    def stderr(self) -> str:
        output_file = Path(self.work_directory, "logs", "error.log")
        return output_file.read_text()

    def config(self) -> dict[str, typing.Any]:
        cfg: dict[str, typing.Any] = {
            "command": self.command,
        }
        if self.wall_time:
            cfg["wall_time"] = self.wall_time
        if self.ntasks > 0:
            cfg["ntasks"] = self.ntasks
        return cfg

    def batch_file(self) -> str:
        "Get the content of the batch file submited to the batch manager."
        return json.dumps(self.config())

    def _prepare_run(self) -> None:
        Path(self.work_directory).mkdir(parents=True, exist_ok=True)
        for fi in self.input_files:
            copy(fi, self.work_directory)
        if self.ntasks > 0:
            nodelist = (socket.gethostname() + "\n") * self.ntasks
            nodefile = Path(self.work_directory) / "batch_nodefile.txt"
            nodefile.write_text(nodelist)
        log_path = Path(self.work_directory) / "logs"
        log_path.mkdir(parents=True, exist_ok=True)

    def _posix_submit(self) -> None:
        r_pipe, w_pipe = os.pipe()
        pid = os.fork()
        if pid > 0:
            # father side - get pid of the grand child
            self.pid = int(os.fdopen(r_pipe).readline().strip())
            return
        # double fork for daemon creation
        try:
            pid = os.fork()
            if pid > 0:
                # child side
                try:
                    os.write(w_pipe, f"{pid}\n".encode())
                finally:
                    os._exit(0)  # ensure child ends here!
            # actual daemon
            self._run()
        finally:
            os._exit(0)  # ensure grand child ends here!

    def _run(self) -> None:
        # daemonize process
        os.setsid()
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        log_path = Path(self.work_directory) / "logs"
        stdout_log = log_path / "output.log"
        stderr_log = log_path / "error.log"
        # file descriptors are automaticaly closed by default
        # (see close_fds argument of Popen).
        stdout_file = open(stdout_log, "w")
        stderr_file = open(stderr_log, "w")
        proc = subprocess.Popen(
            self.command,
            cwd=self.work_directory,
            stdout=stdout_file,
            stderr=stderr_file,
        )
        signal.signal(signal.SIGTERM, functools.partial(handler, proc))
        try:
            exit_code = proc.wait(self.wall_time)
        except subprocess.TimeoutExpired:
            proc.terminate()
            exit_code = proc.wait()
        exit_log = log_path / "exit_code.log"
        with open(exit_log, "w") as exit_file:
            exit_file.write(str(exit_code))

    def _windows_submit(self) -> None:
        # Windows implementation : never join the child.
        # flags DETACHED_PROCESS & CREATE_NEW_PROCESS_GROUP are windows specific
        log_path = Path(self.work_directory) / "logs"
        stdout_log = log_path / "output.log"
        stderr_log = log_path / "error.log"
        # check rights
        stdout_log.touch()
        stderr_log.touch()

        command = [sys.executable, "-m", "pybatch.plugins.local.executor"]
        if self.wall_time is None:
            wall_time = "0"
        else:
            wall_time = repr(self.wall_time)
        command.append(wall_time)
        command.extend(self.command)
        proc = subprocess.Popen(
            command,
            cwd=self.work_directory,
            creationflags=subprocess.DETACHED_PROCESS  # type: ignore
            | subprocess.CREATE_NEW_PROCESS_GROUP,  # type: ignore
        )
        self.pid = proc.pid



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
import typing

import pybatch
import pybatch.protocols.local

import tempfile
from pathlib import Path
import shutil
import os
import sys

import tests.job_cases


def local_case_config(
    plugin: str, config: dict[str, typing.Any], case_name: str, script_name: str
) -> tuple[pybatch.LaunchParameters, pybatch.GenericProtocol, str]:
    if "work_dir" in config:
        work_dir = os.path.join(config["work_dir"], case_name + "_" + plugin)
    else:
        work_dir = tempfile.mkdtemp(suffix="_pybatchtest")
    params = pybatch.LaunchParameters([], work_dir, python_exe=sys.executable)
    if "wckey" in config:
        params.wckey = config["wckey"]
    params.ntasks = 1

    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / script_name
    params.input_files = [script]
    protocol = pybatch.protocols.local.LocalProtocol()
    return params, protocol, work_dir


def test_hello(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "hello", "hello.py"
    )
    tests.job_cases.test_hello(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_sleep(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "sleep", "sleep.py"
    )
    tests.job_cases.test_sleep(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_cancel(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "cancel", "sleep.py"
    )
    tests.job_cases.test_cancel(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_error(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "error", "error.py"
    )
    tests.job_cases.test_error(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_nodefile(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "nodefile", "check_nodefile.py"
    )
    tests.job_cases.test_nodefile(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_reconnect(
    local_plugin: str, local_args: dict[str, typing.Any]
) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "reconnect", "sleep.py"
    )
    tests.job_cases.test_reconnect(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_array(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "array", "array.py"
    )
    tests.job_cases.test_array(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_array_ko(local_plugin: str, local_args: dict[str, typing.Any]) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "array_ko", "array_ko.py"
    )
    tests.job_cases.test_array_ko(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)


def test_array_cancel(
    local_plugin: str, local_args: dict[str, typing.Any]
) -> None:
    job_params, protocol, work_dir = local_case_config(
        local_plugin, local_args, "array_cancel", "array.py"
    )
    tests.job_cases.test_array_cancel(local_plugin, protocol, job_params)
    shutil.rmtree(work_dir)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
import typing
from pathlib import Path
import os

import pybatch
import pybatch.tools

import tests.job_cases


def create_launch_parameters(
    config: dict[str, typing.Any],
) -> pybatch.LaunchParameters:
    params = pybatch.LaunchParameters([], config["work_dir"])
    if "wckey" in config:
        params.wckey = config["wckey"]
    if "is_posix" in config:
        params.is_posix = config["is_posix"]
    if "python_exe" in config:
        params.python_exe = config["python_exe"]
    params.wall_time = "1"  # one minute
    return params


def create_protocol(
    protocol_name: str, config: dict[str, typing.Any]
) -> pybatch.GenericProtocol:
    import pybatch

    params = pybatch.ConnectionParameters(config["host"])
    if "user" in config:
        params.user = config["user"]
    if "password" in config:
        params.password = config["password"]
    if "gss_auth" in config:
        params.gss_auth = config["gss_auth"]
    if protocol_name == "ssh":
        import pybatch.protocols.ssh

        protocol = pybatch.protocols.ssh.SshProtocol(params)
    elif protocol_name == "paramiko":
        import pybatch.protocols.paramiko

        protocol = pybatch.protocols.paramiko.ParamikoProtocol(params)
    else:
        raise Exception(f"Unknown protocol {protocol_name}")
    return protocol


def remote_case_config(
    remote_plugin: str,
    remote_protocol: str,
    remote_args: dict[str, typing.Any],
    case_name: str,
    script_name: str,
) -> tuple[pybatch.LaunchParameters, pybatch.GenericProtocol]:
    is_posix = True
    if "is_posix" in remote_args:
        is_posix = remote_args["is_posix"]
    job_params = create_launch_parameters(remote_args)
    case_dir = case_name + "_" + remote_plugin + "_" + remote_protocol
    job_params.work_directory = pybatch.tools.path_join(
        job_params.work_directory,
        case_dir,
        is_posix=is_posix,
    )
    job_params.ntasks = 1
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / script_name

    job_params.input_files = [script]
    protocol = create_protocol(remote_protocol, remote_args)
    return job_params, protocol


def clean(protocol: pybatch.GenericProtocol, remote_dir: str) -> None:
    pycommand = f'import shutil; shutil.rmtree("{remote_dir}")'
    protocol.run(["python3", "-c", pycommand])


def test_hello(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "hello", "hello.py"
    )
    tests.job_cases.test_hello(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_sleep(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "sleep", "sleep.py"
    )
    tests.job_cases.test_sleep(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_cancel(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "cancel", "sleep.py"
    )
    tests.job_cases.test_cancel(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_error(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "error", "error.py"
    )
    tests.job_cases.test_error(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_nodefile(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin,
        remote_protocol,
        remote_args,
        "nodefile",
        "check_nodefile.py",
    )
    tests.job_cases.test_nodefile(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_reconnect(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "reconnect", "sleep.py"
    )
    tests.job_cases.test_reconnect(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_array(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "array", "array.py"
    )
    tests.job_cases.test_array(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_array_ko(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "array_ko", "array_ko.py"
    )
    tests.job_cases.test_array_ko(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)


def test_array_cancel(
    remote_plugin: str, remote_protocol: str, remote_args: dict[str, typing.Any]
) -> None:
    job_params, protocol = remote_case_config(
        remote_plugin, remote_protocol, remote_args, "array_cancel", "array.py"
    )
    tests.job_cases.test_array_cancel(remote_plugin, protocol, job_params)
    clean(protocol, job_params.work_directory)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
from collections.abc import Iterable
from pathlib import Path
import paramiko
import scp  # type: ignore
from typing import Any
from ..parameter import ConnectionParameters
from .. import PybatchException
from ..tools import escape_str


class ParamikoProtocol:
    """Communication protocol based on python modules paramiko and scp.

    The ssh connection is opened when the object is created and it is closed
    when the object is garbage collected.
    """

    def __init__(self, params: ConnectionParameters):
        client = paramiko.client.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client = client
        self.params = params
        self._open()

    def __del__(self) -> None:
        self.client.close()

    def _open(self) -> None:
        try:
            self.client.connect(
                self.params.host,
                username=self.params.user,
                password=self.params.password,
                gss_auth=self.params.gss_auth,
            )
        except Exception as e:
            message = f"Failed to open ssh connection to {self.params.host}."
            raise PybatchException(message) from e

    def upload(
        self, local_entries: Iterable[str | Path], remote_path: str
    ) -> None:
        with scp.SCPClient(self.client.get_transport()) as client:
            for entry in local_entries:
                client.put(entry, remote_path, recursive=True)

    def download(
        self, remote_entries: Iterable[str], local_path: str | Path
    ) -> None:
        with scp.SCPClient(self.client.get_transport()) as client:
            for entry in remote_entries:
                client.get(entry, local_path, recursive=True)

    def create(self, remote_path: str, content: str) -> None:
        with self.client.open_sftp() as sftp:
            with sftp.open(remote_path, "w") as remote_file:
                remote_file.write(content)

    def read(self, remote_path: str) -> str:
        with self.client.open_sftp() as sftp:
            with sftp.open(remote_path, "r") as remote_file:
                return remote_file.read().decode()

    def run(self, command: list[str]) -> str:
        # in case of issues with big stdout|stderr see
        # https://github.com/paramiko/paramiko/issues/563
        if len(command) == 0:
            raise PybatchException("Empty command.")
        str_command = command[0]
        for arg in command[1:]:
            str_command += " " + escape_str(arg)
        stdin, stdout, stderr = self.client.exec_command(str_command)
        stdin.close()
        str_std = stdout.read().decode()
        str_err = stderr.read().decode()
        stdout.close()
        stderr.close()
        ret_code = stdout.channel.recv_exit_status()
        if ret_code != 0:
            message = f"""Error {ret_code}.
  command: {str_command}
  server: {self.params.host}
  stderr: {str_err}
"""
            raise PybatchException(message)
        return str_std

    def __getstate__(self) -> dict[str, Any]:
        # deal with paramiko.client which is not suppported by pickle
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state["client"]
        return state

    def __setstate__(self, state: Any) -> None:
        self.__dict__.update(state)
        client = paramiko.client.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client = client
        self._open()


def open(params: ConnectionParameters) -> ParamikoProtocol:
    return ParamikoProtocol(params)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
from importlib.metadata import entry_points
from .genericjob import GenericJob
from .generic_protocol import GenericProtocol
from .parameter import LaunchParameters


def create_job(
    plugin_name: str,
    params: LaunchParameters,
    connection_protocol: GenericProtocol | None = None,
) -> GenericJob:
    """Create the job with the chosen plugin.

    :param plugin_name: name of the plugin to use for the job creation.
    :param params: job parameters.
    :param connection_protocol: protocol for remote connection. None for local
     use.
    """

    # for entry_point in entry_points().get("pybatch.plugins"):
    ep = entry_points()
    if isinstance(ep, dict):
        # older python version
        ep_it = ep["pybatch.plugins"]
    else:
        ep_it = ep.select(group="pybatch.plugins")
    for entry_point in ep_it:
        if entry_point.name == plugin_name:
            plugin = entry_point.load()()
            job: GenericJob = plugin.create_job(params, connection_protocol)
            return job
    raise Exception(f"Plugin {plugin_name} not found.")


def reload_job(dump: str) -> GenericJob:  # type: ignore
    """Reload a job from a dumped string - for future use, not implemented.

    :param dump: representation of the job.
    """
    ...  # TODO



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
import typing
import tempfile
import os
import shutil
from pathlib import Path


def test_protocol_paramiko(remote_args: dict[str, typing.Any]) -> None:
    import pybatch
    import pybatch.protocols.paramiko
    from pybatch.tools import path_join
    import scp

    # Test connection to a host that does not exist
    connect_param = pybatch.ConnectionParameters(host="noname_zozo")
    try:
        p = pybatch.protocols.paramiko.ParamikoProtocol(connect_param)
    except pybatch.PybatchException as e:
        assert str(e) == "Failed to open ssh connection to noname_zozo."
    else:
        assert 0  # Exception expected
    # Test configuration
    hostname = remote_args["host"]
    work_dir = remote_args["work_dir"]
    gss_auth = False
    if "gss_auth" in remote_args:
        gss_auth = remote_args["gss_auth"]
    is_posix = True
    if "is_posix" in remote_args:
        is_posix = remote_args["is_posix"]
    connect_param = pybatch.ConnectionParameters(
        host=hostname, gss_auth=gss_auth
    )
    if "python_exe" in remote_args:
        python_exe = remote_args["python_exe"]
    else:
        python_exe = "python3"
    if "user" in remote_args:
        connect_param.user = remote_args["user"]
    if "password" in remote_args:
        connect_param.password = remote_args["password"]
    p = pybatch.protocols.paramiko.ParamikoProtocol(connect_param)
    local_work_dir = tempfile.mkdtemp(suffix="_pybatchtest")
    test_file_name = "paramiko_test.txt"
    remote_test_file = path_join(work_dir, test_file_name, is_posix=is_posix)
    local_test_file = os.path.join(local_work_dir, test_file_name)
    # Test download a remote path that does not exist.
    try:
        p.download([remote_test_file], local_test_file)
    except scp.SCPException as e:
        # The remote path should be included in the error message
        # but the full message may depend on the language.
        assert remote_test_file in str(e)
    else:
        assert 0

    # Test create a file in an inaccessible place
    try:
        p.create("/no/directory/", "file content")
    except FileNotFoundError:
        pass
    else:
        assert 0

    # Test create + download
    file_content = "Servus!"
    p.create(remote_test_file, file_content)
    remote_content = p.read(remote_test_file)
    assert remote_content == file_content
    p.download([remote_test_file], local_test_file)
    assert Path(local_test_file).read_text() == file_content

    # download in an inaccessible place
    local_wrong_path = os.path.join(local_work_dir, "nodir", "nofile")
    try:
        p.download([remote_test_file], local_wrong_path)
    except FileNotFoundError:
        pass
    else:
        assert 0

    # upload nonexistent file
    try:
        p.upload([local_wrong_path], work_dir)
    except FileNotFoundError as e:
        # assert compatible Linux & Windows
        assert "nodir" in str(e)
        assert "nofile" in str(e)
    else:
        assert 0

    # upload file to an inaccessible place
    try:
        p.upload([local_test_file], "/no/directory/")
    except scp.SCPException as e:
        assert "/no/directory/" in str(e)
    else:
        assert 0

    # upload + download + check
    name_bis = "paramiko_test_bis.txt"
    remote_test_file_bis = path_join(work_dir, name_bis, is_posix=is_posix)
    local_test_file_bis = os.path.join(local_work_dir, name_bis)
    p.upload([local_test_file], remote_test_file_bis)
    p.download([remote_test_file_bis], local_test_file_bis)
    assert Path(local_test_file_bis).read_text() == file_content

    # run
    command = [python_exe, "-c", 'print("Cool!")']
    res = p.run(command)
    assert res.strip() == "Cool!"

    # run error
    command = [python_exe, "-c", "exit(1)"]
    try:
        res = p.run(command)
    except pybatch.PybatchException as e:
        assert "Error 1" in str(e)
        assert f"command: {python_exe} -c 'exit(1)'" in str(e)
    else:
        assert 0

    # remove remote files
    pycommand = f'import os; os.remove("{remote_test_file}")'
    p.run([python_exe, "-c", pycommand])
    pycommand = f'import os; os.remove("{remote_test_file_bis}")'
    p.run([python_exe, "-c", pycommand])

    shutil.rmtree(local_work_dir)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
from pathlib import Path
import time

from pybatch import GenericJob, GenericProtocol, LaunchParameters
from pybatch import PybatchException
from pybatch.protocols.local import LocalProtocol

from pybatch.tools import path_join, is_absolute, escape_str


def simplified_state(name: str) -> str:
    finished_states = ["COMPLETED"]
    running_states = ["CONFIGURI", "RUNNING"]
    queued_states = ["PENDING"]
    paused_states = ["RESV_DEL_", "REQUEUE", "RESIZING", "SUSPENDED"]
    failed_states = [
        "BOOT_FAIL",
        "CANCELLED",
        "DEADLINE",
        "FAILED",
        "NODE_FAIL",
        "OUT_OF_ME",
        "PREEMPTED",
        "REVOKED",
        "SIGNALING",
        "SPECIAL_E",
        "STAGE_OUT",
        "STOPPED",
        "TIMEOUT",
    ]
    for state in finished_states:
        if name.startswith(state):
            return "FINISHED"
    for state in running_states:
        if name.startswith(state):
            return "RUNNING"
    for state in queued_states:
        if name.startswith(state):
            return "QUEUED"
    for state in paused_states:
        if name.startswith(state):
            return "PAUSED"
    for state in failed_states:
        if name.startswith(state):
            return "FAILED"
    return ""


def reduce_states(states_list: list[str], max_number_of_states: int) -> str:
    simple_states = [simplified_state(s) for s in states_list]
    if "RUNNING" in simple_states:
        return "RUNNING"  # at least one is running
    if "QUEUED" in simple_states:
        return "QUEUED"  # at least one is queued and none is running
    if "PAUSED" in simple_states:
        return "PAUSED"
    if max_number_of_states == len(states_list):
        # if all the jobs are listed by squeue (not always the case!)
        if "FAILED" in simple_states:
            return "FAILED"
        else:
            return "FINISHED"
    return ""


class Job(GenericJob):
    def __init__(
        self, param: LaunchParameters, protocol: GenericProtocol | None
    ):
        self.job_params = param
        self.protocol: GenericProtocol
        if protocol is None:
            self.protocol = LocalProtocol()
        else:
            self.protocol = protocol
        self.jobid = ""
        self.number_of_jobs = self.job_params.total_jobs

    def submit(self) -> None:
        """Submit the job to the batch manager and return.

        If the submission fails, raise an exception.
        """
        # with self.protocol as protocol:
        try:
            # create remote workdir
            # workdir is always a linux path
            logdir = path_join(
                self.job_params.work_directory, "logs", is_posix=True
            )
            command = ["mkdir", "-p", logdir]
            self.protocol.run(command)

            batch_path = path_join(
                self.job_params.work_directory, "batch.cmd", is_posix=True
            )
            self.protocol.create(batch_path, self.batch_file())
            self.protocol.upload(
                self.job_params.input_files, self.job_params.work_directory
            )
            output = self.protocol.run(
                [
                    "sbatch",
                    "--parsable",
                    "--chdir",
                    self.job_params.work_directory,
                    batch_path,
                ]
            )
            self.jobid = output.split(";")[0].strip()
            int(self.jobid)  # check
            self.number_of_jobs = self.job_params.total_jobs
        except Exception as e:
            message = "Failed to submit job."
            raise PybatchException(message) from e

    def wait(self) -> None:
        "Wait until the end of the job."
        if not self.jobid:
            return
        state = self.state()
        while state != "FINISHED" and state != "FAILED":
            time.sleep(1)
            state = self.state()

    def state(self) -> str:
        """Possible states : 'CREATED', 'QUEUED', 'RUNNING',
        'PAUSED', 'FINISHED', 'FAILED'
        """
        if not self.jobid:
            return "CREATED"
        try:
            # with self.protocol as protocol:
            # First try to query the job with "squeue" command
            try:
                command = ["squeue", "-h", "-o", "%T", "-j", self.jobid]
                squeue_state = self.protocol.run(command)
                if squeue_state:
                    st = ""
                    if self.number_of_jobs > 1:
                        list_states = squeue_state.splitlines()
                        st = reduce_states(list_states, self.number_of_jobs)
                    else:
                        st = simplified_state(squeue_state)
                    if st:
                        return st
            except PybatchException:
                # job was finished a long time ago and it is no longer
                # available for squeue
                pass

            # If "squeue" failed, the job may be finished.
            # In this case, try to query the job with "sacct".
            command = [
                "sacct",
                "-X",  # ignore steps
                "-o",  # output fields
                "State%-10",  # state field on less than 10 chars
                "-n",  # no header
                "-j",  # jobid
                self.jobid,
            ]
            sacct_state = self.protocol.run(command)
            max_tries = 5
            while not sacct_state and max_tries:
                # if not sacct_state:
                # Give some time to slurm scheduler to update
                max_tries -= 1
                time.sleep(1)
                sacct_state = self.protocol.run(command)
            st = ""
            if self.number_of_jobs > 1:
                list_states = sacct_state.splitlines()
                st = reduce_states(list_states, self.number_of_jobs)
                if not st:
                    # No RUNNING, no PENDING and
                    # len(list_states) < number_of_jobs
                    # The main job in the array is the job which is launched the
                    # last but its state is PENDING from the start of the array
                    # until the job is actually launched. The other jobs are not
                    # listed from the start. They are listed only when they can
                    # be launched, when the jobs that are before in the array
                    # are finished.
                    # When a job is scheduled to be launched, it takes a little
                    # while before seeing it with the acct command. This is why
                    # it is possible to have all listed jobs FINISHED or FAILED,
                    # but some jobs of the array not listed yet.
                    if "CANCELLED" in sacct_state:
                        # The main job is not "PENDING" and there are less
                        # states listed than the total number of jobs in the
                        # array. At least one job has cancelled state.
                        # This happens when the job array is cancelled, but also
                        # when a particular job in the array is cancelled after
                        # the end of the main job, if the main job is shorter
                        # than other jobs.
                        # WARNING If the main job has started and an individual
                        # job in the array was cancelled, the arrays is set as
                        # FAILED despite some jobs may be still running but not
                        # yet scheduled. To be investigated.
                        st = "FAILED"
                    else:
                        # The main job finished very fast and the scheduler have
                        # not added to queue all the jobs of the last slice in
                        # the array yet.
                        st = "RUNNING"
            else:
                st = simplified_state(sacct_state)
        except Exception as e:
            raise PybatchException("Failed to get the state of the job.") from e
        if st:
            return st
        else:
            raise PybatchException(
                f"Unknown state. squeue_state: {squeue_state}, sacct_state:{sacct_state}"
            )

    def exit_code(self) -> int | None:
        if not self.jobid:
            return None
        state = self.state()
        if state == "FINISHED":
            return 0
        if state != "FAILED":
            return None
        try:
            command = [
                "sacct",
                # "-X",  # ignore steps
                "-o",  # output fields
                "ExitCode%-10",
                "-n",  # no header
                "-j",  # jobid
                self.jobid,
            ]
            # code_str format: <exit_code>:<signal_received>
            # If ok, code_str is "0:0"
            # If cancel, code_str is "0:15"
            # If exit 1, code_str is "1:0"

            code_str = self.protocol.run(command)
            result = None
            if code_str:
                # The result is the first non 0 value found, job steps included.
                result = 0
                for val in code_str.splitlines():
                    signal_received = int(val.split(":")[1])
                    if signal_received != 0:
                        cur_val = signal_received
                    else:
                        cur_val = int(val.split(":")[0])
                    if cur_val != 0:
                        result = cur_val
                        break
        except Exception:
            result = None
        return result

    def cancel(self) -> None:
        "Stop the job."
        if not self.jobid:
            return
        command = ["scancel", self.jobid]
        try:
            self.protocol.run(command)
        except Exception as e:
            raise PybatchException("Failed to cancel the job.") from e

    def get(self, remote_paths: list[str], local_path: str | Path) -> None:
        """Copy a file or directory from the remote work directory.

        :param remote_paths: paths relative to work directory on remote host.
        :param local_path: destination of the copy on local file system.
        """
        checked_paths = []
        for path in remote_paths:
            if is_absolute(path, self.job_params.is_posix):
                checked_paths.append(path)
            else:
                p = path_join(
                    self.job_params.work_directory,
                    path,
                    is_posix=self.job_params.is_posix,
                )
                checked_paths.append(p)
        self.protocol.download(checked_paths, local_path)

    def stdout(self) -> str:
        output_file = path_join(
            self.job_params.work_directory,
            "logs",
            "output.log",
            is_posix=self.job_params.is_posix,
        )
        return self.protocol.read(str(output_file))

    def stderr(self) -> str:
        output_file = path_join(
            self.job_params.work_directory,
            "logs",
            "error.log",
            is_posix=self.job_params.is_posix,
        )
        return self.protocol.read(str(output_file))

    def batch_file(self) -> str:
        "Get the content of the batch file submited to the batch manager."
        batch = """#!/bin/bash -l
#SBATCH --output=logs/output.log
#SBATCH --error=logs/error.log
"""
        if self.job_params.name:
            batch += f"#SBATCH --job-name={self.job_params.name}\n"
        if self.job_params.total_jobs > 1:
            simul = ""
            if self.job_params.max_simul_jobs > 1:
                simul = f"%{self.job_params.max_simul_jobs}"
            array = f"0-{self.job_params.total_jobs - 1}{simul}"
            batch += f"#SBATCH --array={array}\n"
        if self.job_params.ntasks > 0:
            batch += f"#SBATCH --ntasks={self.job_params.ntasks}\n"
        if self.job_params.nodes > 0:
            batch += f"#SBATCH --nodes={self.job_params.nodes}\n"
        if self.job_params.exclusive:
            batch += "#SBATCH --exclusive\n"
        if self.job_params.wall_time:
            batch += f"#SBATCH --time={self.job_params.wall_time}\n"
        if self.job_params.mem_per_node:
            batch += f"#SBATCH --mem={self.job_params.mem_per_node}\n"
        if self.job_params.mem_per_cpu:
            batch += f"#SBATCH --mem-per-cpu={self.job_params.mem_per_cpu}\n"
        if self.job_params.queue:
            batch += f"#SBATCH --qos={self.job_params.queue}\n"
        if self.job_params.partition:
            batch += f"#SBATCH --partition={self.job_params.partition}\n"
        if self.job_params.wckey:
            batch += f"#SBATCH --wckey={self.job_params.wckey}\n"
        for extra in self.job_params.extra_as_list:
            batch += f"#SBATCH {extra}\n"
        if self.job_params.extra_as_string:
            batch += self.job_params.extra_as_string
        if self.job_params.create_nodefile:
            batch += """
LIBBATCH_NODEFILE=`pwd`/batch_nodefile.txt
srun hostname > $LIBBATCH_NODEFILE
export LIBBATCH_NODEFILE
"""
        # batch += "echo Jobid: $SLURM_JOB_ID\n"
        # if self.job_params.total_jobs > 1:
        #     batch += "echo master Jobid: $SLURM_ARRAY_JOB_ID\n"
        command = self.job_params.command
        str_command = command[0]
        for arg in command[1:]:
            str_command += " " + escape_str(arg)
        if self.job_params.total_jobs > 1:
            str_command += " $SLURM_ARRAY_TASK_ID"
        batch += "\n"
        batch += str_command
        return batch

    # A réfléchir, mais il vaut peut-être mieux utiliser la sérialisation
    # pickle.
    # def dump(self) -> str:
    # " Serialization of the job."
    # ...



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
def test_version():
    import pybatch

    assert pybatch.__version__ == "1.0"


def test_plugin():
    import pybatch

    params = pybatch.LaunchParameters("script", "dir")
    import pybatch.plugins.local.job

    job = pybatch.create_job("local", params)
    assert isinstance(job, pybatch.plugins.local.job.Job)

    import pybatch.plugins.slurm.job

    job = pybatch.create_job("slurm", params)
    assert isinstance(job, pybatch.plugins.slurm.job.Job)

    import pybatch.plugins.nobatch.job

    job = pybatch.create_job("nobatch", params)
    assert isinstance(job, pybatch.plugins.nobatch.job.Job)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
from __future__ import annotations
import typing
from collections.abc import Iterable
from pathlib import Path


class GenericProtocol(typing.Protocol):
    """Connection protocol (ssh, local, ...).

    This interface defines the services expected from a connection protocol.
    """

    def upload(
        self, local_entries: Iterable[str | Path], remote_path: str
    ) -> None:
        "Upload files and directories to the server."
        ...

    def download(
        self, remote_entries: Iterable[str], local_path: str | Path
    ) -> None:
        "Download files and directories from the server."
        ...

    def create(self, remote_path: str, content: str) -> None:
        "Create a file on the server."
        ...

    def read(self, remote_path: str) -> str:
        "Get the content of a file."
        ...

    def run(self, command: list[str]) -> str:
        "Run a command on the server."
        ...



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
import pybatch.plugins.nobatch.pybatch_manager as manager

import tempfile
from pathlib import Path
import os
import shutil
import inspect
import subprocess
import time
import sys


def test_hello():
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "hello.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)

    # submit
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        py_exe,
        "hello.py",
        "zozo",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # wait
    args = [py_exe, manager_script, "wait", pid]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FINISHED"
    output_file = Path(workdir) / "logs" / "output.log"
    assert output_file.read_text().strip() == "Hello zozo !"

    # clean
    shutil.rmtree(workdir)


def test_sleep():
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)

    # submit long job
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        py_exe,
        "sleep.py",
        "1",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    # wait
    args = [py_exe, manager_script, "wait", pid]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FINISHED"
    result_file = Path(workdir) / "wakeup.txt"
    assert result_file.exists()

    # clean
    shutil.rmtree(workdir)


def test_cancel():
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)

    # submit long job
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        py_exe,
        "sleep.py",
        "1",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    # cancel
    args = [py_exe, manager_script, "cancel", pid]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0

    time.sleep(2)

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FAILED"
    time.sleep(2)
    result_file = Path(workdir) / "wakeup.txt"
    assert not result_file.exists()

    # clean
    shutil.rmtree(workdir)


def test_timeout():
    """Wall time shorter than execution time."""
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)

    # submit long job
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        "--wall_time",
        "1",
        py_exe,
        "sleep.py",
        "3",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    time.sleep(2)

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FAILED"
    time.sleep(2)
    result_file = Path(workdir) / "wakeup.txt"
    assert not result_file.exists()

    # clean
    shutil.rmtree(workdir)


def test_notimeout():
    """Walltime longer than execution time."""
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "sleep.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)

    # submit
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        "--wall_time",
        "3",
        py_exe,
        "sleep.py",
        "1",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    time.sleep(2)

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FINISHED"
    result_file = Path(workdir) / "wakeup.txt"
    assert result_file.exists()

    # clean
    shutil.rmtree(workdir)


def test_array():
    "Simulation of a job array Slurm"
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "array.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        "--total_jobs",
        "4",
        py_exe,
        "array.py",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    # wait
    args = [py_exe, manager_script, "wait", pid]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FINISHED"
    for idx in range(4):
        result = Path(workdir) / f"result_{idx}.txt"
        assert result.read_text() == str(idx)

    # clean
    shutil.rmtree(workdir)


def test_array_ko():
    "Job array with a failed job"
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "array_ko.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        "--total_jobs",
        "4",
        py_exe,
        "array_ko.py",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    # wait
    args = [py_exe, manager_script, "wait", pid]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0

    # check
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FAILED"
    for idx in range(4):
        result = Path(workdir) / f"result_{idx}.txt"
        assert result.read_text() == str(idx)
    exit_code = Path(workdir) / "logs" / "exit_code.log"
    assert "42" == exit_code.read_text()

    # clean
    shutil.rmtree(workdir)


def test_array_cancel():
    "Cancel on a job array."
    py_exe = sys.executable
    workdir = tempfile.mkdtemp(suffix="_pybatchtest")
    current_file_dir = os.path.dirname(__file__)
    script = Path(current_file_dir) / "scripts" / "array.py"
    shutil.copy(script, workdir)
    manager_script = shutil.copy(inspect.getfile(manager), workdir)
    args = [
        py_exe,
        manager_script,
        "submit",
        workdir,
        "--total_jobs",
        "4",
        py_exe,
        "array.py",
    ]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    pid = proc.stdout.strip()
    assert int(pid) > 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "RUNNING"

    # cancel
    args = [py_exe, manager_script, "cancel", pid]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0

    # state
    args = [py_exe, manager_script, "state", pid, workdir]
    proc = subprocess.run(args, capture_output=True, text=True)
    assert proc.returncode == 0
    state = proc.stdout.strip()
    assert state == "FAILED"

    # clean
    shutil.rmtree(workdir)



# Copyright (C) 2025  CEA, EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
"""
This script manages the execution of a command.
It is meant to be run on the remote server side.
This script should avoid the utilisation of non standard python modules for
better compatibility, because pybatch is not necessary installed on the remote
server.
"""

# No typing for better compatibility with older python versions.
# (edf gaia - python 3.5)
# mypy: ignore-errors
import argparse
import subprocess
import signal
import functools
import os
from pathlib import Path
import sys
import logging
import socket

global interrupted
interrupted = False

try:
    import psutil

    def process_exists(pid):
        return psutil.pid_exists(pid)

    def process_kill(pid):
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        for p in children:
            try:
                p.terminate()
            except psutil.NoSuchProcess:
                pass
        parent.terminate()

except ModuleNotFoundError:
    # Implementation if psutil is not available on the server side.
    def process_exists(pid):
        proc_exists = True
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            proc_exists = False
        return proc_exists

    def process_kill(pid):
        os.kill(pid, signal.SIGTERM)


def handler(proc, signum, frame):
    proc.terminate()
    global interrupted
    interrupted = True


def run_one_job(workdir, command, wall_time, log_path, stdout_log, stderr_log):
    # file descriptors are automaticaly closed by default
    # (see close_fds argument of Popen).
    stdout_file = open(stdout_log, "w")  # python 3.5
    stderr_file = open(stderr_log, "w")  # python 3.5
    message = "Launch command: " + str(command)
    logging.info(message)
    proc = subprocess.Popen(
        command,
        stdout=stdout_file,
        stderr=stderr_file,
        cwd=workdir,
        creationflags=0,
    )
    signal.signal(signal.SIGTERM, functools.partial(handler, proc))
    try:
        exit_code = proc.wait(wall_time)
    except subprocess.TimeoutExpired:
        proc.terminate()
        logging.info("Timeout expired! Terminate child.")
        exit_code = proc.wait()
    exit_log = str(log_path / "exit_code.log")  # python 3.5
    with open(exit_log, "w") as exit_file:
        exit_file.write(str(exit_code))


def run_many_jobs(
    workdir,
    command,
    wall_time,
    log_path,
    stdout_log,
    stderr_log,
    total_jobs,
    max_simul_jobs,
):
    # TODO deal with max_simul_jobs
    global_exit_code = 0
    for idx in range(total_jobs):
        # file descriptors are automaticaly closed by default
        # (see close_fds argument of Popen).
        stdout_file = open(stdout_log, "a")  # python 3.5
        stderr_file = open(stderr_log, "a")  # python 3.5
        current_command = command + [str(idx)]
        message = "Launch command[{}]: {}"
        message = message.format(idx, current_command)
        logging.info(message)
        proc = subprocess.Popen(
            current_command, stdout=stdout_file, stderr=stderr_file, cwd=workdir
        )
        signal.signal(signal.SIGTERM, functools.partial(handler, proc))
        try:
            exit_code = proc.wait(wall_time)
        except subprocess.TimeoutExpired:
            proc.terminate()
            message = "Timeout expired! Terminate child[{}].".format(idx)
            logging.info(message)
            exit_code = proc.wait()
        if exit_code != 0:
            global_exit_code = exit_code
        global interrupted
        if interrupted:
            break
    exit_log = str(log_path / "exit_code.log")  # python 3.5
    with open(exit_log, "w") as exit_file:
        exit_file.write(str(global_exit_code))


def run(
    workdir,
    wall_time,
    total_jobs,
    max_simul_jobs,
    command,
):
    log_path = Path(workdir) / "logs"
    stdout_log = str(log_path / "output.log")
    stderr_log = str(log_path / "error.log")
    manager_log = str(log_path / "manager.log")
    if not wall_time:
        wall_time = None

    log = open(manager_log, "w")  # python 3.5
    # std redirection
    os.dup2(log.fileno(), sys.stdout.fileno())
    os.dup2(log.fileno(), sys.stderr.fileno())
    os.dup2(os.open(os.devnull, os.O_RDWR), sys.stdin.fileno())

    if total_jobs > 1:
        run_many_jobs(
            workdir,
            command,
            wall_time,
            log_path,
            stdout_log,
            stderr_log,
            total_jobs,
            max_simul_jobs,
        )
    else:
        run_one_job(
            workdir, command, wall_time, log_path, stdout_log, stderr_log
        )


def posix_submit(
    workdir,
    wall_time,
    total_jobs,
    max_simul_jobs,
    command,
):
    pid = os.fork()
    if pid > 0:
        # father side
        logging.info("Jobid " + str(pid))
        print(pid)
        return
    # child side
    os.setsid()
    signal.signal(signal.SIGHUP, signal.SIG_IGN)
    run(workdir, wall_time, total_jobs, max_simul_jobs, command)


def windows_submit(
    workdir,
    wall_time,
    total_jobs,
    max_simul_jobs,
    command,
):
    "Used for windows, when fork is not available."
    current_script = __file__
    py_exe = sys.executable
    if wall_time is None:
        wall_time = 0
    run_command = [
        py_exe,
        current_script,
        "run",
        workdir,
        str(wall_time),
        str(total_jobs),
        str(max_simul_jobs),
    ]
    run_command.extend(command)
    logging.info(str(run_command))
    stdout_file = open(
        str(Path(workdir) / "logs" / "submit_std.log"), "w"
    )  # python 3.5
    stderr_file = open(
        str(Path(workdir) / "logs" / "submit_err.log"), "w"
    )  # python 3.5
    proc = subprocess.Popen(
        run_command,
        cwd=workdir,
        stdout=stdout_file,
        stderr=stderr_file,
        # Windows specific flags.
        creationflags=subprocess.DETACHED_PROCESS
        | subprocess.CREATE_NEW_PROCESS_GROUP,
    )
    print(proc.pid)


def submit(workdir, command, wall_time, ntasks, total_jobs, max_simul_jobs):
    """Launch a command and return immediatly.

    The command is launched in workdir.
    The pid of the created process is printed on stdout.
    """
    message = (
        "submit workdir={}, command={}, walltime={}, ntasks={}, total_jobs={}"
    )
    message = message.format(workdir, command, wall_time, ntasks, total_jobs)
    logging.info(message)
    # check access to workdir before fork.
    log_path = Path(workdir) / "logs"
    log_path.mkdir(parents=True, exist_ok=True)
    stdout_log = log_path / "output.log"
    stderr_log = log_path / "error.log"
    manager_log = log_path / "manager.log"
    stdout_log.touch()
    stderr_log.touch()
    manager_log.touch()
    # generate batch_nodefile.txt
    if ntasks > 0:
        nodelist = (socket.gethostname() + "\n") * ntasks
        nodefile = Path(workdir) / "batch_nodefile.txt"
        nodefile.write_text(nodelist)

    # execute in detached mode
    if hasattr(os, "fork"):
        posix_submit(workdir, wall_time, total_jobs, max_simul_jobs, command)
    else:
        windows_submit(workdir, wall_time, total_jobs, max_simul_jobs, command)


def wait(proc_id):
    "Wait for the process to finish."
    logging.info("Wait jobid " + str(proc_id))
    import time

    while process_exists(proc_id):
        time.sleep(0.1)
    logging.info("Wait finished.")


def state(proc_id, workdir):
    """Print the state of the process.

    The work directory of the job is supposed to be the same as the directory
    of this script.
    """
    logging.info("state jobid " + str(proc_id))
    proc_exists = process_exists(proc_id)
    if proc_exists:
        logging.info("state RUNNING")
        print("RUNNING")
    else:
        exit_log = Path(workdir) / "logs" / "exit_code.log"
        if exit_log.is_file():
            exit_value = exit_log.read_text().strip()
            if exit_value == "0":
                logging.info("state FINISHED")
                print("FINISHED")
            else:
                logging.info("state FAILED")
                print("FAILED")
        else:
            logging.info("state FAILED")
            print("FAILED")


def cancel(proc_id):
    "Kill the process."
    logging.info("cancel jobid " + str(proc_id))
    try:
        process_kill(proc_id)
    except Exception:
        # TODO
        logging.info("cancel kill failed!")


def main(args_list=None):
    parser = argparse.ArgumentParser(description="Job manager for pybatch.")
    subparsers = parser.add_subparsers(
        dest="mode",  # required=True,#python>=3.7
        help="Use mode.",
    )

    # SUBMIT
    parser_submit = subparsers.add_parser(
        "submit", help="Start a job and print the pid."
    )
    parser_submit.add_argument("work_dir", help="Work directory.")
    parser_submit.add_argument(
        "--wall_time",
        type=int,
        default=None,
        help="Maximum execution time in seconds.",
    )
    parser_submit.add_argument(
        "--ntasks",
        type=int,
        default=0,
        help="If positive, generate batch_nodefile.txt.",
    )
    parser_submit.add_argument(
        "--total_jobs",
        type=int,
        default=1,
        help="Number of jobs to launch, 1 by default.",
    )
    parser_submit.add_argument(
        "--max_simul_jobs",
        type=int,
        default=1,
        help="For future use. Parameter ignored.",
    )
    parser_submit.add_argument("command", nargs="+", help="Command to submit.")

    # RUN
    parser_run = subparsers.add_parser("run", help="Run in background.")
    parser_run.add_argument("work_dir", help="Work directory.")
    parser_run.add_argument(
        "wall_time",
        type=int,
        help="Maximum execution time in seconds.",
    )
    parser_run.add_argument(
        "total_jobs",
        type=int,
        help="Number of jobs to launch.",
    )
    parser_run.add_argument(
        "max_simul_jobs",
        type=int,
        help="For future use. Parameter ignored.",
    )
    parser_run.add_argument("command", nargs="+", help="Command to submit.")

    # WAIT
    parser_wait = subparsers.add_parser(
        "wait", help="Wait for the end of a job."
    )
    parser_wait.add_argument("proc", type=int, help="Process id.")

    # STATE
    parser_state = subparsers.add_parser(
        "state", help="Print the state of a job."
    )
    parser_state.add_argument("proc", type=int, help="Process id.")
    parser_state.add_argument("work_dir", help="Work directory.")

    # CANCEL
    parser_cancel = subparsers.add_parser("cancel", help="Cancel a job.")
    parser_cancel.add_argument("proc", type=int, help="Process id.")

    args = parser.parse_args(args_list)
    if args.mode == "submit":
        submit(
            args.work_dir,
            args.command,
            args.wall_time,
            args.ntasks,
            args.total_jobs,
            args.max_simul_jobs,
        )
    elif args.mode == "wait":
        wait(args.proc)
    elif args.mode == "state":
        state(args.proc, args.work_dir)
    elif args.mode == "cancel":
        cancel(args.proc)
    elif args.mode == "run":
        run(
            args.work_dir,
            args.wall_time,
            args.total_jobs,
            args.max_simul_jobs,
            args.command,
        )
    else:
        print("No command defined!")


def log_config():
    workdir = os.path.dirname(sys.argv[0])
    log_file = os.path.join(workdir, "pybatch.log")
    logging.basicConfig(
        filename=log_file,
        level=logging.DEBUG,
        format="%(asctime)s - %(message)s",
    )


if __name__ == "__main__":
    log_config()
    main()
